(      Title:  A utility for testing Forth
               floating-point stack effects
        File:  ftester.fs
  Test files:  ttester-xf-test.fs, ttester-xfarray-test.fs 
    Log file:  ttester-xf.log
     License:  John Hayes, Public Domain
     Version:  1.2.1
     Revised:  November 29, 2010
  Revised by:  David N. Williams

For any code derived from John Hayes' tester program:
)
\  (C) 1995 JOHNS HOPKINS UNIVERSITY / APPLIED PHYSICS LABORATORY
\  MAY BE DISTRIBUTED FREELY AS LONG AS THIS COPYRIGHT NOTICE REMAINS.
(
The rest is public domain.

This file revises the code for ttester, a utility for testing
Forth words, as developed by several authors.  See the file
ttester-xf.fs, version 1.2.0 or later, for history.

  http://tinyurl.com/ttester-xf-102

This file extracts the part of ttester-xf 1.0.0 that deals with
the separated floating-point stack, hereafter called the fp
stack, into an independent module called ftester.  The code
assumes that floating-point support is present, and the file
will not load unles the fp stack is separated, as required by
the Forth 200x standard.  The part that deals with the data
stack has been extracted into an independent xtester module. 
Versions 1.0.2 and later of ttester-xf load both xtester and
ftester, and remain drop-in compatible with ttester, unless the
user redefines error reporting by using the ttester words
ERROR-XT or ERROR1.

Version 1.2.0 adds new fp comparisons, based on suggestions by
Krishna Myneni in comp.lang.forth:

  http://tinyurl.com/ttester-xf-102

Words not declared as public are considered to be private, not
guaranteed to be available in future versions of ftester.

PUBLIC WORDS

   tests:  F{  F->  }F
           FT-ABS-ERROR  FT-REL-ERROR  FT-AREL-ERROR
           FT-TEST=-XT
           SET-FT-MODE-EXACT  SET-FT-MODE-ABS
           SET-FT-MODE-REL    SET-FT-MODE-AREL
           
           for ttester compatibiity
           ABS-NEAR   REL-NEAR
           SET-EXACT  SET-NEAR

  errors:  FT-ERROR-XT  FT-ERROR-DEFAULT  FT-ERROR-INDEX

  arrays:  FT-INITIALS  FT-RESULTS  FT-GOALS
           FT-A@  FT-A!  FT-A<>

  config:  MAX-FT-INITIALS  MAX-FT-RESULTS

The array words can be used for more elaborate error reporting,
for example, to dump the fp results and goals when something is
wrong, or to report a particular fp error with the help of
FT-ERROR-INDEX.  See the word FT-A. in ttester-xf-test.fs, and
an example of its use to print fp results and goals.

The basic test usage takes the form:

  <finitials> F{ <fresults> F-> <fgoals> }F <finitials>

where <finitials> are possibly nonempty fp stack contents at the
start of the test.  The sequence saves <finitials> and removes
them from the fp stack, executes the code that produces the fp
stack <fresults>, saves and removes the <fresults> from the fp
stack, executes the code that produces the fp stack <fgoals>,
saves and removes them, compares the saved <fresults> and
<fgoals>, reports when there is a discrepancy between the two,
and restores the saved <finitials>.

For example:

F{ 1E 2E fswap F-> 2E 1E }F ok
F{ 1E 2E fswap F-> 2E 2E }T INCORRECT FP RESULT: F{ 1E 2E fswap F-> 2E 2E }F ok
F{ 1E 2E fswap F-> 2E }F WRONG NUMBER OF FP RESULTS: F{ 1E 2E fswap F-> 2E }F ok

Effects present or produced on non-fp stacks before F{, F->, or
}F are simply ignored.

More examples can be found in the file ttester-xf-test.fs.

The private word FT-ERROR is vectored.  It executes the xt held
in the public variable FT-ERROR-XT, which defaults to the xt of
the public FT-ERROR-DEFAULT.  For example, the user might change
the action to count the number of fp errors for later reporting,
and use FT-ERRROR-DEFAULT as a factor, which prints a string and
displays the line on which the error occurs.

Floating-point testing can involve further complications.  There
are the perennial issues of fp value comparisons.  Exact equality
is specified by SET-FT-MODE-EXACT [the default].  If one of the
included approximate equality tests is desired, execute
SET-FT-MODE-ABS, SET-FT-MODE-REL, or SET-FT-MODE-AREL.
Corresponding to these, the public fvariables FT-ABS-ERROR
[default 0E], FT-REL-ERROR [default 1E-12], and FT-AREL-ERROR
[defaulty 1E-12] contain the values to be used in comparisons by
the private words FT-ABS=, FT-REL=, and FT-AREL=.  Here is an
example:

  SET-FT-MODE-REL
  1E-6 FT-REL-ERROR F!
  F{ S" 3.14159" >FLOAT F-> -1E FACOS TRUE }F

Note that comparisona in this mode are not symmetric.  Measured
values must be in the fresults, and reference values in the
fgoals.

The user may supply a different equality test by storing
his own xt in the public variable FT-TEST=-XT.

Storage of <finitials>, <fresults>, and <fgoals> includes
overflow checking.  The user can change the overflow limits by
defining one or both of the following constants *before* loading
this file:

  MAX-FT-INITIALS  MAX-FT-RESULTS

The default xt's and tolerances in the various public variables
and fvariables can be changed after loading this file.

Loading ftester.fs does not change BASE.  Remember that
floating-point input is ambiguous if the base is not decimal.

Ftester uses the same limited set of non-CORE words as the fp
stack section of ttester, plus [UNDEFINED], and 2R>.  See
ttester-xf.fs for the list.
)

BASE @
DECIMAL

DEPTH 1E DEPTH 1- FDROP = 0=
[IF] .( FLOATING-STACK REQUIRED) ABORT [THEN]


\ *** ERROR REPORTING
(
Error reporting in ttester-xf is unavoidably different from that
in ttester, because ttester-xt does not use a common execution
vector for data and fp stack reports.  That of course does not
affect compatibility as long as no errors occur.
)

\ PUBLIC

VARIABLE FT-ERROR-XT
: FT-ERROR   FT-ERROR-XT @ EXECUTE ;

VARIABLE FT-ERROR-INDEX  \ holds array index|-1 after error

: FT-ERROR-DEFAULT   ( c-addr u -- )
(
Display an error message followed by the line that had the
error.
)
  TYPE SOURCE TYPE CR ;
' FT-ERROR-DEFAULT FT-ERROR-XT !


\ *** TTESTER-XF FP COMPARISONS
(
The public words in this section provide an alternative to the
ttester fp equality tests.  The ttester-xf default is the same
exact equality as that of ttester.
)

\ PUBLIC

VARIABLE FT-TEST=-XT

\ PRIVATE

: FT-TEST=  ( f: x y -- ) ( -- flag )  FT-TEST=-XT @ EXECUTE ;

\ PUBLIC

\ The sign of these tolerances must be plus.
FVARIABLE FT-ABS-ERROR      0E FT-ABS-ERROR  F!
FVARIABLE FT-REL-ERROR   1E-12 FT-REL-ERROR  F!
FVARIABLE FT-AREL-ERROR  1E-12 FT-AREL-ERROR F!

\ PRIVATE

: FT-DATUM=  ( f: x y -- ) ( -- flag )
(
Leave TRUE if the two floats have the same internal
representation, including the IEEE-FP 2008 special data, signed
NAN with load, and signed zero and infinity.  Else leave FALSE.

Whether the specials work with F~ is implementation dependent,
according to DPANS94.
)
  0E F~ ;

\ |x - y| < eps
: FT-ABS=  ( f: x y -- ) ( -- flag )  FT-ABS-ERROR F@ F~ ;

\ |m - r| < eps * |r|
: FT-REL=  ( f: meas ref -- ) ( -- flag )
  FSWAP FOVER F- FSWAP F/ FABS FT-REL-ERROR F@ F< ;

\ |x - y| < eps * (|x| + |y|)/2 
: FT-AREL= ( f: x y -- ) ( -- flag )  FT-AREL-ERROR F@ FNEGATE 2E F/ F~ ;

\ PUBLIC

: SET-FT-MODE-EXACT ( -- )  ['] FT-DATUM= FT-TEST=-XT ! ;
: SET-FT-MODE-ABS   ( -- )  ['] FT-ABS=   FT-TEST=-XT ! ;
: SET-FT-MODE-REL   ( -- )  ['] FT-REL=   FT-TEST=-XT ! ;
: SET-FT-MODE-AREL  ( -- )  ['] FT-AREL=  FT-TEST=-XT ! ;

SET-FT-MODE-EXACT


\ *** TTESTER FP COMPARISONS
(
Exact and approximate fp equality in ttester-xf remains
compatible with ttester, if the public words in this section are
used instead of the newer ttester-xf words.
)

\ PUBLIC
(
Set the following to the relative and absolute tolerances you
want for approximate float equality, to be used with F~ in
FT-NEARLY=.  Keep the signs, because F~ needs them.
)
  FVARIABLE REL-NEAR   1E-12 REL-NEAR F!
  FVARIABLE ABS-NEAR      0E ABS-NEAR F!

\ PRIVATE
(
When FT-EXACT? is TRUE, FT-CONF= uses FT-FDATUM=, otherwise
FT-NEARLY=.
)
  TRUE VALUE FT-EXACT?

: FT~ABS=  ( f: x y -- ) ( -- flag )
(
Leave TRUE if the two floats are equal within the tolerance
stored in FT-ABS-NEAR , else FALSE.
)
  ABS-NEAR F@ F~ ;

: FT~REL=  ( f: x y -- ) ( -- flag )
(
Leave TRUE if the two floats are relatively equal based on the
tolerance stored in FT-REL-NEAR , else FALSE.
)
  REL-NEAR F@ FNEGATE F~ ;

[UNDEFINED] F2DUP  [IF] : F2DUP  FOVER FOVER ; [THEN]
[UNDEFINED] F2DROP [IF] : F2DROP FDROP FDROP ; [THEN]

: FT-NEARLY=  ( f: x y -- ) ( -- flag )
(
Leave TRUE if the two floats are nearly equal, else FALSE.  This
is a refinement of Dirk Zoller's FEQ to also allow x = y,
including both zero, or to allow approximately equality when x
and y are too small to satisfy the relative approximation mode
in the F~ specification.
)
  F2DUP FT-DATUM= IF F2DROP TRUE EXIT THEN
  F2DUP FT~REL=     IF F2DROP TRUE EXIT THEN
  FT~ABS= ;

: FT-CONF=  ( f: x y -- ) ( -- flag )
  FT-EXACT? IF FT-DATUM= ELSE FT-NEARLY= THEN ;

\ PUBLIC

: SET-EXACT  ( -- )   TRUE TO FT-EXACT? ['] FT-CONF= FT-TEST=-XT ! ;
: SET-NEAR   ( -- )  FALSE TO FT-EXACT? ['] FT-CONF= FT-TEST=-XT ! ;


\ *** ARRAY WORDS
(
The first array slot has index one, not zero.
)

\ PUBLIC

\ USER CONFIG DEFAULTS
[UNDEFINED] MAX-FT-INITIALS [IF] 32 CONSTANT MAX-FT-INITIALS [THEN]
[UNDEFINED] MAX-FT-RESULTS  [IF] 32 CONSTANT MAX-FT-RESULTS  [THEN]

\ Arrays begin with a cell-sized count.
HERE 1 CELLS ALLOT FALIGN MAX-FT-INITIALS FLOATS ALLOT CONSTANT FT-INITIALS
HERE 1 CELLS ALLOT FALIGN MAX-FT-RESULTS  FLOATS ALLOT CONSTANT FT-RESULTS
HERE 1 CELLS ALLOT FALIGN MAX-FT-RESULTS  FLOATS ALLOT CONSTANT FT-GOALS

\ overflow checking to be done elsewhere
: FT-A!   ( +n addr -- ) ( f: rn ... r1 -- )
  2DUP ! CELL+ FALIGNED SWAP ( f-addr n) DUP
  IF   FLOATS OVER + SWAP
       DO I F! 1 FLOATS +LOOP 
  ELSE 2DROP THEN ;

: FT-A@   ( addr -- +n ) ( f: -- rn ... r1 )
  DUP @ DUP >R ( addr n r: n)
  IF   CELL+ FALIGNED R@ 1- FLOATS OVER + 
       DO I F@ -1 FLOATS +LOOP
  ELSE DROP THEN R> ;

[UNDEFINED] NIP [IF] : NIP  ( a b -- b)  SWAP DROP ; [THEN]

: FT-A<>  ( addr1 addr2 -- index|-1|0 )
  DUP @ SWAP CELL+ FALIGNED >R >R  ( addr1 r: f-addr2 #left )
  DUP @ R@ <> IF 2R> 2DROP DROP -1 EXIT THEN
  ( addr1) DUP CELL+ FALIGNED
  BEGIN  R@ ( addr1 f-addr1 #left)
  WHILE  ( f-addr1) DUP F@ FLOAT+
         R> ( #left) 1- R>
         ( f-addr2) DUP F@ FLOAT+ >R ( #left) >R
         FT-TEST=
  WHILE
  REPEAT ( addr1 f-addr1) DROP @ ( len) 2R> NIP - ( index)
  ELSE   ( addr1 f-addr1) 2DROP 2R> 2DROP 0
  THEN ;


\ *** TEST WORDS

\ PUBLIC

: F{   ( f: <finitials> -- )
(
Save the fp stack <finitials> for restoration by }F.  Clear
<fresults> and <fgoals> storage to put it in a defined state
for }F in case F-> is missing.
)
  FDEPTH DUP MAX-FT-INITIALS > ABORT" TOO MANY FP INITIALS"
  FT-INITIALS FT-A!   0 FT-RESULTS !   0 FT-GOALS ! ;

: F->  ( f: fresults -- )
(
Record the depth and contents of the fp stack.
)
  FDEPTH DUP MAX-FT-RESULTS > ABORT" TOO MANY FP RESULTS"
  FT-RESULTS FT-A! ;

: }F   ( f: fgoals -- finitials )
(
Save the fp stack <fgoals>, compare them with the saved
<fresults>, and restore the <finitials>.  Note that when the
comparison corresponds to SET-FT-MODE-REL, the measured values
must be fresults and the reference values must be fgoals.
)
  FDEPTH DUP MAX-FT-RESULTS > ABORT" TOO MANY FP GOALS"
  FT-GOALS FT-A!
  FT-RESULTS FT-GOALS FT-A<> DUP FT-ERROR-INDEX !
  ?DUP IF -1 =
    IF   S" WRONG NUMBER OF FP RESULTS: "
    ELSE S" INCORRECT FP RESULT: " THEN
    FT-ERROR
  THEN
  FT-INITIALS FT-A@ ( len) DROP ;


BASE !
\ *** END of ftester.fs
