(      Title:  A utility for testing Forth
               data stack effects
        File:  xtester.fs
  Test files:  ttester-xf-test.fs, ttester-xfarray-test.fs 
    Log file:  ttester-xf.log
     License:  John Hayes, Public Domain
     Version:  1.2.1
     Revised:  November 29, 2010
  Revised by:  David N. Williams

For any code derived from John Hayes' tester program:
)
\  (C) 1995 JOHNS HOPKINS UNIVERSITY / APPLIED PHYSICS LABORATORY
\  MAY BE DISTRIBUTED FREELY AS LONG AS THIS COPYRIGHT NOTICE REMAINS.
(
The rest is public domain.

This file revises the code for ttester, a utility for testing
Forth words, as developed by several authors.  See the file
ttester-xf.fs, version 1.2.0 or later, for history.

This file extracts the part of ttester-xf 1.0.0 that deals with
the data stack into an independent module called xtester.  Here
"x" stands for the generic ANS/ISO Forth "unspecified cell" data
type.  The part that deals with the floating-point stack has
been extracted into an independent ftester module.  Versions
1.0.2 and later of ttester-xf load both xtester and ftester, and
remain drop-in compatible with ttester, unless the user
redefines error reporting by using the ttester words ERROR-XT or
ERROR1.

Words not declared as public are considered to be private, not
guaranteed to be available in future versions of xtester.

PUBLIC WORDS
   tests:  X{  X->  }X
  errors:  XT-ERROR-XT  XT-ERROR-DEFAULT  XT-ERROR-INDEX
  arrays:  XT-INITIALS  XT-RESULTS  XT-GOALS
           XT-A@  XT-A!  XT-A<>
  config:  MAX-XT-INITIALS  MAX-XT-RESULTS

The array words can be used for more elaborate error reporting,
for example, to dump results and goals when something is wrong,
or to report a particular error with the help of XT-ERROR-INDEX.
See the word XT-A. in ttester-xf-test.fs, and an example of its
use to print results and goals.

The basic test usage takes the form:

  <xinitials> X{ <xresults> X-> <xgoals> }X <xinitials>

where <xinitials> are possibly nonempty data stack contents at
the start of the test.  The sequence saves <xinitials> and
removes them from the data stack, executes the code that
produces the data stack <xresults>, saves and removes the
<xresults> from the stack, executes the code that produces the
data stack <xgoals>, saves and removes them, compares the saved
<xresults> and <xgoals>, reports when there is a discrepancy
between the two, and restores the saved <xinitials>.

The three X test words have the same basic functionality as the
original Hayes words {, ->, and }.  The key differences are that
they save and restore <xinitials>, save <xgoals>, and have a
user interface for error-reporting.  These have evolved from
ttester and ttester-xf 1.0.0.

A usage example:

X{ 1 2 3 swap X-> 1 3 2 }X ok
X{ 1 2 3 swap X-> 1 2 2 }X INCORRECT XRESULT: X{ 1 2 3 swap X-> 1 2 2 }X ok
X{ 1 2 3 swap X-> 1 2 }X WRONG NUMBER OF XRESULTS: X{ 1 2 3 swap X-> 1 2 }X ok

Effects present or produced on non-data stacks before X{, X->,
or }X are simply ignored.

More examples can be found in the file ttester-xf-test.fs.

The private word XT-ERROR is vectored.  It executes the xt held
in the public variable XT-ERROR-XT, which defaults to the xt of
the public XT-ERROR-DEFAULT.  For example, the user might change
the action to count the number of errors for later reporting,
and use XT-ERRROR-DEFAULT as a factor, which prints a string and
displays the line on which the error occurs.

Storage of <xinitials>, <xresults>, and <xgoals> includes
overflow checking.  The user can change the overflow limits by
defining one or both of the following CONSTANT's *before*
loading this file:

  MAX-XT-INITIALS  MAX-XT-RESULTS

The xt in the variable XT-ERROR-XT can be changed only *after*
loading this file.

Loading xtester.fs does not change BASE.

Xtester uses the same limited set of non-CORE words as the data
stack section of ttester, plus [UNDEFINED], 2>R, and 2R>.  See
ttester-xf.fs for the list.
)
BASE @
DECIMAL


\ *** ERROR REPORTING

VARIABLE XT-ERROR-XT
: XT-ERROR XT-ERROR-XT @ EXECUTE ;  \ vectored error reporting

VARIABLE XT-ERROR-INDEX  \ holds array index|-1 after error

: XT-ERROR-DEFAULT   ( c-addr u -- )
(
Display an error message followed by the line that had the
error.
)
  TYPE SOURCE TYPE CR ;
' XT-ERROR-DEFAULT XT-ERROR-XT !


\ *** ARRAY WORDS
(
The first array slot has index one, not zero.
)

\ USER CONFIG DEFAULTS
[UNDEFINED] MAX-XT-INITIALS [IF] 32 CONSTANT MAX-XT-INITIALS [THEN]
[UNDEFINED] MAX-XT-RESULTS  [IF] 32 CONSTANT MAX-XT-RESULTS  [THEN]

\ Arrays begin with a cell-sized count.
CREATE XT-INITIALS MAX-XT-INITIALS 1+ CELLS ALLOT
CREATE XT-RESULTS  MAX-XT-RESULTS  1+ CELLS ALLOT
CREATE XT-GOALS    MAX-XT-RESULTS  1+ CELLS ALLOT

\ overflow checking to be done elsewhere
: XT-A!  ( xn ... x1 +n addr -- )
  OVER 1+ CELLS OVER + SWAP DO I ! 1 CELLS +LOOP ;

: XT-A@  ( addr -- xn ... x1 +n )
  DUP @ CELLS OVER + DO I @ -1 CELLS +LOOP ;

[UNDEFINED] NIP [IF] : NIP  ( a b -- b)  SWAP DROP ; [THEN]

: XT-A<>  ( addr1 addr2 -- index|-1|0 )
  DUP @ 2>R    ( addr1 r: addr2 #left )
  DUP @ R@ <> IF 2R> 2DROP DROP -1 EXIT THEN
  ( addr1) DUP
  BEGIN  CELL+ R@    ( addr1 addr1' #left)
  WHILE  R> R> CELL+ ( addr1 addr1' #left addr2')
         DUP >R SWAP 1- >R @ OVER @ =
  WHILE
  REPEAT ( addr1 addr1') DROP @ ( len) 2R> NIP - ( index)
  ELSE   ( addr1 addr1') 2DROP 2R> 2DROP 0
  THEN ;


\ *** TESTS

: X{   ( <xinitials> -- )
(
Save the data stack <xinitials> for restoration by }X.  Clear
<xresults> and <xgoals> storage to put it in a defined state for
}X in case X-> is missing.
)
  DEPTH DUP MAX-XT-INITIALS > ABORT" TOO MANY INITIALS"
  XT-INITIALS XT-A!   0 XT-RESULTS !   0 XT-GOALS ! ;

: X->  ( <xresults> -- )
(
Record the depth and contents of the data stack.
)
  DEPTH DUP MAX-XT-RESULTS > ABORT" TOO MANY RESULTS"
  XT-RESULTS XT-A! ;

: }X   ( <xgoals> -- <xinitials> )
(
Save the data stack <xgoals>, compare them with the saved
<xresults>, and restore the <xinitials>.
)
  DEPTH DUP MAX-XT-RESULTS > ABORT" TOO MANY GOALS"
  XT-GOALS XT-A!
  XT-RESULTS XT-GOALS XT-A<> DUP XT-ERROR-INDEX !
  ?DUP IF -1 =
    IF   S" WRONG NUMBER OF RESULTS: "
    ELSE S" INCORRECT RESULT: " THEN
    XT-ERROR
  THEN
  XT-INITIALS XT-A@ ( len) DROP ;

BASE !
\ *** END of xtester.fs
